@@ -7,8 +7,10 @@ import os
 import sys
 
 
-VERSION = '3.0'
-INSTALL_PREFIX = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
+VERSION = '3.2'
+INSTALL_PREFIX = os.path.normpath(
+    os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
+)
 MIN_PYTHON_VERSION = (3, 6, 0)
 
 # Check python version
