@@ -20,11 +20,14 @@ from reframe.core.logging import getlogger
 
 
 class JobScheduler(abc.ABC):
-    '''Abstract base class for job scheduler backends.'''
+    '''Abstract base class for job scheduler backends.
+
+    :meta private:
+    '''
 
     @abc.abstractmethod
-    def completion_time(self, job):
-        '''The completion time of this job expressed in seconds from the Epoch.
+    def make_job(self, *args, **kwargs):
+        '''Create a new job to be managed by this scheduler.
 
         :meta private:
         '''
@@ -90,6 +93,15 @@ class JobScheduler(abc.ABC):
         :meta private:
         '''
 
+    @abc.abstractmethod
+    def poll(self, *jobs):
+        '''Poll all the requested jobs.
+
+        :arg jobs: The job descriptors to poll.
+
+        :meta private:
+        '''
+
 
 class Job:
     '''A job descriptor.
@@ -142,53 +154,6 @@ class Job:
     #:
     #: :type: :class:`reframe.core.launchers.JobLauncher`
     launcher = fields.TypedField('launcher', JobLauncher)
-    scheduler = fields.TypedField('scheduler', JobScheduler)
-
-    #: .. versionadded:: 2.21
-    #:
-    #: The ID of the current job.
-    #:
-    #: :type: :class:`int` or :class:`None`.
-    jobid = fields.TypedField('jobid', int, type(None))
-
-    #: .. versionadded:: 2.21
-    #:
-    #: The exit code of the job.
-    #:
-    #: This may or may not be set depending on the scheduler backend.
-    #:
-    #: :type: :class:`int` or :class:`None`.
-    exitcode = fields.TypedField('exitcode', int, type(None))
-
-    #: .. versionadded:: 2.21
-    #:
-    #: The state of the job.
-    #:
-    #: The value of this field is scheduler-specific.
-    #:
-    #: :type: :class:`str` or :class:`None`.
-    state = fields.TypedField('state', str, type(None))
-
-    #: .. versionadded:: 2.17
-    #:
-    #: The list of node names assigned to this job.
-    #:
-    #: This attribute is :class:`None` if no nodes are assigned to the job
-    #: yet.
-    #: This attribute is set reliably only for the ``slurm`` backend, i.e.,
-    #: Slurm *with* accounting enabled.
-    #: The ``squeue`` scheduler backend, i.e., Slurm *without* accounting,
-    #: might not set this attribute for jobs that finish very quickly.
-    #: For the ``local`` scheduler backend, this returns an one-element list
-    #: containing the hostname of the current host.
-    #:
-    #: This attribute might be useful in a flexible regression test for
-    #: determining the actual nodes that were assigned to the test.
-    #: For more information on flexible node allocation, see the
-    #: |--flex-alloc-nodes|_ command-line option
-    #:
-    #: This attribute is *not* supported by the ``pbs`` scheduler backend.
-    nodelist = fields.TypedField('nodelist', typ.List[str], type(None))
 
     # The sched_* arguments are exposed also to the frontend
     def __init__(self,
@@ -200,11 +165,6 @@ class Job:
                  max_pending_time=None,
                  sched_flex_alloc_nodes=None,
                  sched_access=[],
-                 sched_account=None,
-                 sched_partition=None,
-                 sched_reservation=None,
-                 sched_nodelist=None,
-                 sched_exclude_nodelist=None,
                  sched_exclusive_access=None,
                  sched_options=None):
 
@@ -216,14 +176,7 @@ class Job:
         self.num_cpus_per_task = None
         self.use_smt = None
         self.time_limit = None
-        self.options = sched_options or []
-
-        # Live job information; to be filled during job's lifetime by the
-        # scheduler
-        self.jobid = None
-        self.exitcode = None
-        self.state = None
-        self.nodelist = None
+        self.options = list(sched_options) if sched_options else []
 
         self._name = name
         self._workdir = workdir
@@ -231,22 +184,30 @@ class Job:
         self._stdout = stdout or '%s.out' % name
         self._stderr = stderr or '%s.err' % name
         self._max_pending_time = max_pending_time
-        self._completion_time = None
 
         # Backend scheduler related information
         self._sched_flex_alloc_nodes = sched_flex_alloc_nodes
         self._sched_access = sched_access
-        self._sched_nodelist = sched_nodelist
-        self._sched_exclude_nodelist = sched_exclude_nodelist
-        self._sched_partition = sched_partition
-        self._sched_reservation = sched_reservation
-        self._sched_account = sched_account
         self._sched_exclusive_access = sched_exclusive_access
 
+        # Live job information; to be filled during job's lifetime by the
+        # scheduler
+        self._jobid = None
+        self._exitcode = None
+        self._state = None
+        self._nodelist = None
+        self._submit_time = None
+        self._completion_time = None
+
+        # Job errors discovered while polling; if not None this will be raised
+        # in finished()
+        self._exception = None
+
     @classmethod
     def create(cls, scheduler, launcher, *args, **kwargs):
-        ret = Job(*args, **kwargs)
-        ret.scheduler, ret.launcher = scheduler, launcher
+        ret = scheduler.make_job(*args, **kwargs)
+        ret._scheduler = scheduler
+        ret.launcher = launcher
         return ret
 
     @property
@@ -281,33 +242,88 @@ class Job:
     def sched_access(self):
         return self._sched_access
 
-    @property
-    def sched_nodelist(self):
-        return self._sched_nodelist
-
-    @property
-    def sched_exclude_nodelist(self):
-        return self._sched_exclude_nodelist
-
-    @property
-    def sched_partition(self):
-        return self._sched_partition
-
-    @property
-    def sched_reservation(self):
-        return self._sched_reservation
-
-    @property
-    def sched_account(self):
-        return self._sched_account
-
     @property
     def sched_exclusive_access(self):
         return self._sched_exclusive_access
 
     @property
     def completion_time(self):
-        return self.scheduler.completion_time(self) or self._completion_time
+        return self._completion_time
+
+    @property
+    def scheduler(self):
+        return self._scheduler
+
+    @property
+    def exception(self):
+        return self._exception
+
+    @property
+    def jobid(self):
+        '''The ID of this job.
+
+        .. versionadded:: 2.21
+
+        .. versionchanged:: 3.2
+           Job ID type is now a string.
+
+        :type: :class:`str` or :class:`None`
+        '''
+        return self._jobid
+
+    @property
+    def exitcode(self):
+        '''The exit code of this job.
+
+        This may or may not be set depending on the scheduler backend.
+
+        .. versionadded:: 2.21
+
+        :type: :class:`int` or :class:`None`
+        '''
+        return self._exitcode
+
+    @property
+    def state(self):
+        '''The state of this job.
+
+        The value of this field is scheduler-specific.
+
+        .. versionadded:: 2.21
+
+        :type: :class`str` or :class:`None`
+        '''
+        return self._state
+
+    @property
+    def nodelist(self):
+        '''The list of node names assigned to this job.
+
+        This attribute is :class:`None` if no nodes are assigned to the job
+        yet.
+        This attribute is set reliably only for the ``slurm`` backend, i.e.,
+        Slurm *with* accounting enabled.
+        The ``squeue`` scheduler backend, i.e., Slurm *without* accounting,
+        might not set this attribute for jobs that finish very quickly.
+        For the ``local`` scheduler backend, this returns an one-element list
+        containing the hostname of the current host.
+
+        This attribute might be useful in a flexible regression test for
+        determining the actual nodes that were assigned to the test.
+        For more information on flexible node allocation, see the
+        |--flex-alloc-nodes|_ command-line option
+
+        This attribute is *not* supported by the ``pbs`` scheduler backend.
+
+        .. versionadded:: 2.17
+
+        :type:`List[str]` or :type:`None`
+        '''
+        return self._nodelist
+
+    @property
+    def submit_time(self):
+        return self._submit_time
 
     def prepare(self, commands, environs=None, **gen_opts):
         environs = environs or []
@@ -349,17 +365,18 @@ class Job:
             return self.sched_flex_alloc_nodes * num_tasks_per_node
 
         available_nodes = self.scheduler.allnodes()
-        getlogger().debug('flex_alloc_nodes: total available nodes %s ' %
+        getlogger().debug('flex_alloc_nodes: total available nodes: %s ' %
                           len(available_nodes))
 
         # Try to guess the number of tasks now
         available_nodes = self.scheduler.filternodes(self, available_nodes)
-        if self.sched_flex_alloc_nodes == 'idle':
+        if self.sched_flex_alloc_nodes.casefold() != 'all':
             available_nodes = {n for n in available_nodes
-                               if n.is_available()}
+                               if n.in_state(self.sched_flex_alloc_nodes)}
             getlogger().debug(
-                'flex_alloc_nodes: selecting idle nodes: '
-                'available nodes now: %s' % len(available_nodes)
+                f'flex_alloc_nodes: selecting nodes in state '
+                f'{self.sched_flex_alloc_nodes!r}: '
+                f'available nodes now: {len(available_nodes)}'
             )
 
         return len(available_nodes) * num_tasks_per_node
@@ -390,6 +407,12 @@ class Job:
 
         return done
 
+    def __eq__(self, other):
+        return type(self) == type(other) and self.jobid == other.jobid
+
+    def __hash__(self):
+        return hash(self.jobid)
+
 
 class Node(abc.ABC):
     '''Abstract base class for representing system nodes.
@@ -398,5 +421,10 @@ class Node(abc.ABC):
     '''
 
     @abc.abstractmethod
-    def is_available(self):
-        '''Return ``True`` if this node is available, ``False`` otherwise.'''
+    def in_state(self, state):
+        '''Returns whether the node is in the given state.
+
+           :arg state: The node state.
+           :returns: :class:`True` if the nodes's state matches the given one,
+                     :class:`False` otherwise.
+        '''
